
class HTTPServiceUnavailableException(Exception):
    pass


class TaskModelMismatchException(Exception):
    pass


class APICallException(Exception):
    pass


class InsufficientParametersException(Exception):
    pass