__title__ = 'hugging_py_face'
__package_name__ = 'hugging_py_face'
__version__ = '0.5.3'
__description__ = "Hugging-Py-Face, the Python client for the Hugging Face Inference API."
__email__ = "minurapunchihewa17@gmail.com"
__author__ = 'Minura Punchihewa'
__github__ = 'https://github.com/MinuraPunchihewa/hugging_py_face'
__pypi__ = 'https://pypi.org/project/hugging-py-face/'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright (C) 2023  Minura Punchihewa'